package POE::Interface;

use strict;

use Carp;
use Data::Dumper;
use XML::DOM;

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(
    load
);
$VERSION = '0.01';

sub DEBUG () { 1 };
sub WARN  () { 1 };

##############################################################################
# Constructors
sub new
{
    my $package=shift;
    my $self=bless {}, $package;
    $self->load(@_) if @_;
    return $self
}

##########################################################
sub load
{
    my($self, $file)=@_;
    if(@_==1) {
        $file=$self;
        $self=__PACKAGE__->new();
        DEBUG and warn "Building ".__PACKAGE__;
    } elsif(not ref $self) {
        DEBUG and warn "Building $self";
        $self=$self->new();
    }
    
    $self->{file}=$file;
    $self->{parser}=XML::DOM::Parser->new(KeepCDATA =>1);
    $self->{xml}=$self->{parser}->parsefile($file);
    die "Unable to parse $file" unless $self->{xml};
    return $self;    
}

##############################################################################
# Helper routines to access the DOM
# While few of these actually need $self, they are methods so that
# they can be overloaded
# Hmm.... i seem to have completely encapsulated the DOM here, which
# kind of makes it pointless.

##########################################################
sub sessions
{
    my($self)=@_;
    die Dumper $self unless $self->{xml};
    return $self->{xml}->getElementsByTagName('SESSION');
}

##########################################################
sub get_session
{
    my($self, $id)=@_;
    foreach my $session ($self->sessions) {
        foreach ($session->getElementsByTagName('PACKAGE')) {
            return $session if $_ eq $id;
        }
        foreach ($session->getElementsByTagName('ALIAS')) {
            return $session if $_ eq $id;
        }
    }
    WARN and warn __PACKAGE__, " Unable to find session $id\n";
    return;
}


##########################################################
sub events
{
    my($self, $session)=@_;
    $session=$self->get_session($session) unless ref $session;
    return unless $session;
    return $session->getElementsByTagName('EVENT');
}

##########################################################
sub get_event
{
    my($self, $session, $name)=@_;
    foreach my $event ($self->events($session)) {
        foreach ($event->getElementsByTagName('NAME')) {
            return $event if $_ eq $name;
        }
    }
    WARN and warn __PACKAGE__, " Unable to find event $name\n";
    return;
}

##########################################################
sub get_tag ($$$:$)
{
    my($self, $node, $name, $n)=@_;
    DEBUG and warn "Looking for tag $name\n";
    return $node->getElementsByTagName($name) if wantarray;
    return $node->getElementsByTagName($name)->item($n||0);
    WARN and warn "Tag $name not found\n";
}

##########################################################
sub get_tag_text ($$$)
{
    my($self, $node, $name)=@_;

    my $tag=$node->getElementsByTagName($name);
    return unless $tag and @$tag;
    $tag=$tag->[0];
    return unless $tag;
    return $tag->getFirstChild->getData;
}

##########################################################
sub get_text ($$)
{
    my($self, $parent)=@_;

    return '' unless defined $parent;
    my $text='';
    my $type;
    foreach my $node ($parent->getChildNodes) {
        $type=$node->getNodeName;
        DEBUG and warn "$type\n";
        next unless $type =~ m(^#(?:cdata-section|text));
        $text .= $node->getData;
    }
    return $text;
}


##########################################################
sub get_attribute
{
    my($self, $node, $name, $default)=@_;

    DEBUG and warn "Looking for attribute $name\n";

    # We use getAttributeNode instead of getAttribute because the later
    # must return '' even if the attribute isn't present.
    # http://www.w3.org/TR/REC-DOM-Level-1/level-one-core.html
    my $a=$node->getAttributeNode($name);

    return $default if not defined $a or            # no node == no attrib
                       not $a->isSpecified;         # no value
    return $a->getValue;
}

#############################################################################
# Parameter validation code generation
sub generate_validation
{
    my($self, $node, $statename)=@_;

    my($code, $name, $type, $required, $default, $fetch, $q, $r);

    $q=0;

    my @nodes=$node->getElementsByTagName('VARIABLE');
    my $num=@nodes;
    if(lc($self->get_attribute($node, 'variables-by', 'generated')) eq 'implementation') {
        warn qq(<EVENT variables-by="implementation"> means the implementation will deal with the event parameters, but there are $num <VARIABLE> tag(s)\n)
            if $num;
        return "#none\n"
    }

    $code=<<PERL;
unless(ref(\$__ARGS) eq 'HASH' ?  $num == keys %\$__ARGS :  $num == @\$__ARGS) {
        warn "Wrong number of parameters for state $statename\\n";
        return;
    }
PERL

    foreach my $param (@nodes) {

        $name=$self->get_attribute($param, 'name');
        die "Attribute name is required for tag <VARIABLE>\n" 
            unless $name;
        die "Name '$name' is not a valid perl identifier\n"
            unless $name=~/^[^\W\d]\w+$/;

        $type=$self->get_attribute($param, 'type', 'scalar');

        $required=$self->get_attribute($param, 'required', 'no');
        $required=0 unless $required =~ m/^(?:y|yes|1|)$/i;
        
        $fetch=qq((ref \$__ARG eq 'HASH' ? \$__ARG->{$name} : \$__ARG->[$q]));
        $r='';
        $r=qq(if(not defined $fetch) { warn "Required parameter $name missing for $statename\\n"; return}\n)
            if $required;

        $q++;
          ################################################################
        if($type eq 'scalar') {
            $code .= <<PERL;
    $r
    my \$$name = $fetch;
    if(ref \$$name) {
        warn "Parameter $name must be a scalar for $statename\\n"; 
        return;
    }
PERL
        } ################################################################
        elsif($type eq 'array') {
            $code .= <<PERL;
    $r
    my \@$name;
    {   my \$t=$fetch;
        unless(ref \$t eq 'ARRAY') {
            warn "Parameter $name must be an array reference for $statename\\n";
            return;
        }
        \@$name = \@\$t;
    }
PERL
        } ################################################################
        elsif($type eq 'hash') {
            $code .= <<PERL;
    $r
    my %$name;
    {   my \$t=$fetch;
        unless(ref \$t eq 'HASH') {
            warn "Parameter $name must be an hash reference for $statename\\n";
            return;
        }
        %$name=\$t;
    }
PERL
        } ################################################################
        elsif($type eq 'array_or_scalar') {
            $code .= <<PERL;
    $r
    my \@$name;
    {   my \$t=$fetch;
        if((ref(\$t)||'ARRAY') ne 'ARRAY') {
            warn "Parameter $name must be an array reference or scalar for $statename\\n";
            return;
        }
        \@$name = ref \$t ? \@\$t : \$t;
    }
PERL
        } ################################################################
        elsif($type eq 'hash_or_array') {
            $code .= <<PERL;
    $r
    my \%$name;
    {   my \$t=$fetch;
        if(ref(\$t) eq 'ARRAY') {
            \@$name\{\@\$t}=(1) x scalar \@\$t;
        } elsif(ref(\$t) eq 'HASH') {
            \%$name = %\$t;
        } else {
            warn "Parameter $name must be an array or hash reference for $statename\\n";
            return;
        }
    }
PERL
        } ################################################################
        elsif($type eq 'hash_or_array_or_scalar') {
            $code .= <<PERL;
    $r
    my \%$name;
    {   my \$t=$fetch;
        if(not ref(\$t)) {
            $name\{\$t} = 1;
        } elsif(ref(\$t) eq 'ARRAY') {
            \@$name\{\@\$t}=(1) x scalar \@\$t;
        } elsif(ref(\$t) eq 'HASH') {
            \%$name = %\$t;
        } else {
            warn "Parameter $name must be an array or hash reference or a scalar for $statename\\n";
            return;
        }
    }
PERL
        } 
        else {
            warn "Unknown type for <VARIABLE> '$type'\n";
        }
    }
    return $code;
}

############################################################################
# POE::Session code generation
sub generate_session
{
    my($self, $session)=@_;
    my $sn = $session;
    $session = $self->get_session($session) unless ref $session;
    croak "Session $sn not found" unless $session;    

    my $code='';
        
    ###########################
    ## Set up package and alias
    my $package=$self->get_text($self->get_tag($session, 'PACKAGE'));
    my $alias  =$self->get_text($self->get_tag($session, 'ALIAS'));
    if($package) {
        $code .= qq(package $package;\nuse strict;\n);
        unless($alias) {
            $alias=$package;
            $alias=~s/\W+/_/g;
        }
    }
    # packages we will need
    $code .="use POE qw(Session);\nuse Carp;\n\n";

    my $breaker='#' x 70;
    
    ###########################
    # code that goes in front of events, for defining static functions, etc...
    my $usercode = $self->get_text($self->get_tag($session, 'PRE-CODE'));
    $code .= "$breaker\n$usercode\n" if($usercode);

    ###########################
    # now do all the events
    my(@events, @published);
    my ($name, $publish, $params, $start);
    foreach my $event ($self->events($session)) {

        $name=$self->get_text($self->get_tag($event, 'NAME'));
        die "<EVENT> tag requires a <NAME> tag\n" 
            unless $name;
        die "Name '$name' is not a valid perl identifier\n"
            unless $name=~/^[^\W\d]\w+$/;
        
        ###########################
        # is this a published event ?
        $publish=$self->get_attribute($event, 'publish', 'no');
        if($publish =~ m(^(?:y|yes|ok|1)$)i) {
            push @published, $name;
        }
        push @events, $name;
        
        ###########################
        # now generate the code
        $usercode = $self->get_text($self->get_tag($event, 'CODE'));
        next unless $usercode;
        $params = $self->generate_validation($event, $name);

        if($name eq '_start') {
            warn "<EVENT> _start defined more then once, using last instance"
                if $start;
            $start=$usercode;
        } else {
            $code .=<<PERL;
$breaker
sub $name 
{
    my \$__ARGS;
    if(defined \$_[ARG1] or ref(\$_[ARG0]) ne 'HASH') {
        \$__ARGS = [\@_[ARG0..\$#_]];
    } else {
        \$__ARGS = \$_[ARG0];
    }
    $params

    my \$heap=\$_[HEAP];
    my \$__sub=sub {$usercode};
    my \@ret=eval {&\$__sub} if wantarray;
    my \$ret=eval {&\$__sub} unless wantarray;
    if(\$@) {
        warn \$@;
        return;
    }
    return \@ret if wantarray;
    return \$ret;
}
    
PERL
        }

    }
    local $"=' ';
                                
    ###########################
    # put in the _start event
    $start||='';
    $code .=<<PERL;
$breaker
sub _start
{
    my(\$heap, \$kernel)=\@_[HEAP, KERNEL];
    \$kernel->alias_set(q($alias));
    $start
PERL

    ###########################
    # we need to add IKC event publishing
    $code .= qq(    \$kernel->call('IKC', 'publish', q($alias), [qw(@published)]);\n)
            if @published;
    $code .= qq(}\n);

    ###########################
    # now we create the session
    push @events, '_start' unless $start;
    $code .=<<PERL;

$breaker
POE::Session->new (__PACKAGE__, 
    [qw(@events)]
);

1;
PERL

    ###########################
    # And some extra things at the bottom
    my $now = localtime;
    $code .=<<EXTRA;

__DATA__
Generated-on: $now
Generated-from: $self->{file}
POE::Interface: $VERSION
XML::DOM: $XML::DOM::VERSION
XML::Parser: $XML::Parser::VERSION
EXTRA
    
    return $code unless wantarray;
    return ($package, $code);
}

############################################################################
# POE::Component::IKC::ClientLite stub generation
sub generate_lite
{
    my($self, $session)=@_;
    my $sn = $session;
    $session = $self->get_session($session) unless ref $session;
    croak "Session $sn not found" unless $session;    

    my $code='';
        
    ###########################
    ## Set up package and alias
    my $package=$self->get_text($self->get_tag($session, 'PACKAGE'));
    my $alias=$self->get_text($self->get_tag($session, 'ALIAS'));
    if($package) {
        $code .= qq(package $package\::Lite;\nuse strict;\n);
        unless($alias) {
            $alias=$package;
            $alias=~s/\W+/_/g;
        }
    }
    # packages we will need
    $code .="use POE::Component::IKC::ClientLite;\nuse Carp;\n\n";

    my $breaker='#' x 70;
    
    ###########################
    # now do all the events
    my(@events);
    my ($name, $publish, $params, $start);
    foreach my $event ($self->events($session)) {

        $name=$self->get_text($self->get_tag($event, 'NAME'));
        die "<EVENT> tag requires a <NAME> tag\n" 
            unless $name;
        die "Name '$name' is not a valid perl identifier\n"
            unless $name=~/^[^\W\d]\w+$/;
        
        ###########################
        # is this a published event ?
        $publish=$self->get_attribute($event, 'publish', 'no');
        next unless ($publish =~ m(^(?:y|yes|ok|1)$)i);
        push @events, $name;
        
        ###########################
        # now generate the code
        $params = $self->generate_validation($event, $name);
        $code .=<<PERL;
$breaker
sub $name 
{
    __open() or croak \$POE::Component::IKC::ClientLite::error;
    my \$__ARGS={\@_};
    $params
    \$poe->post(q($alias/$name), [\@_]);
}
    
PERL
    }

    ###########################
    my $constructor=$self->get_text($self->get_tag($session, 'IKC:Lite'));
    $code .=$breaker."\n".<<'PERL';
require Exporter;
use vars qw($poe @EXPORT_OK @ISA);
@ISA=qw(Exporter);

sub __open
{
    return 1 if $poe;
    require POE::Component::IKC::ClientLite;
    $poe=POE::Component::IKC::ClientLite::create_ikc_client(
PERL
    $code .= $constructor if $constructor;
    $code .= <<'PERL';
    );
    return unless $poe;
    return 1 unless $INC{'Apache.pm'};
    return 1 unless Apache->request;
    Apache->request->register_cleanup(\&_close);
    return 1;
}

sub __close
{
    return unless $poe;
    $poe->disconnect;
    warn "Closing POE connection";
    undef($poe);
}

END 
{
    $poe->disconnect;
}
PERL


    ###########################
    local $"=' ';
    $code .=<<PERL;
\@EXPORT_OK=qw(@events);

1;

__DATA__
PERL

    ###########################
    # And some extra things at the bottom
    my $now = localtime;
    $code .=<<EXTRA;

__DATA__
Generated-on: $now
Generated-from: $self->{file}
POE::Interface: $VERSION
XML::DOM: $XML::DOM::VERSION
XML::Parser: $XML::Parser::VERSION
EXTRA
    
    return $code unless wantarray;
    return ($package.'::Lite', $code);
}

############################################################################
# POE::Session stub generation
sub generate_stubs
{
    my($self, $session)=@_;
    my $sn = $session;
    $session = $self->get_session($session) unless ref $session;
    croak "Session $sn not found" unless $session;    

    my $code='';
        
    ###########################
    ## Set up package and alias
    my $package=$self->get_text($self->get_tag($session, 'PACKAGE'));
    my $alias=$self->get_text($self->get_tag($session, 'ALIAS'));
    if($package) {
        $code .= qq(package $package\::Lite;\nuse strict;\n);
        unless($alias) {
            $alias=$package;
            $alias=~s/\W+/_/g;
        }
    }
    # packages we will need
    $code .="use POE::Component::IKC::ClientLite;\nuse Carp;\n\n";

    my $breaker='#' x 70;
    
    ###########################
    # now do all the events
    my(@events);
    my ($name, $publish, $params, $start);
    foreach my $event ($self->events($session)) {

        $name=$self->get_text($self->get_tag($event, 'NAME'));
        die "<EVENT> tag requires a <NAME> tag\n" 
            unless $name;
        die "Name '$name' is not a valid perl identifier\n"
            unless $name=~/^[^\W\d]\w+$/;
        
        ###########################
        # is this a published event ?
        $publish=$self->get_attribute($event, 'publish', 'no');
        next unless ($publish =~ m(^(?:y|yes|ok|1)$)i);
        push @events, $name;
        
        ###########################
        # now generate the code
        $params = $self->generate_validation($event, $name);
        $code .=<<PERL;
$breaker
sub $name 
{
    __open() or croak \$POE::Component::IKC::ClientLite::error;
    my \$__ARGS={\@_};
    $params
    \$poe->post(q($alias/$name), [\@_]);
}
    
PERL
    }

    ###########################
    my $constructor=$self->get_text($self->get_tag($session, 'IKC:Lite'));
    $code .=$breaker."\n".<<'PERL';
require Exporter;
use vars qw($poe @EXPORT_OK @ISA);
@ISA=qw(Exporter);

sub __open
{
    return 1 if $poe;
    require POE::Component::IKC::ClientLite;
    $poe=POE::Component::IKC::ClientLite::create_ikc_client(
PERL
    $code .= $constructor if $constructor;
    $code .= <<'PERL';
    );
    return unless $poe;
    return 1 unless $INC{'Apache.pm'};
    return 1 unless Apache->request;
    Apache->request->register_cleanup(\&_close);
    return 1;
}

sub __close
{
    return unless $poe;
    $poe->disconnect;
    warn "Closing POE connection";
    undef($poe);
}

END 
{
    $poe->disconnect;
}
PERL


    ###########################
    local $"=' ';
    $code .=<<PERL;
\@EXPORT_OK=qw(@events);

1;

__DATA__
PERL

    ###########################
    # And some extra things at the bottom
    my $now = localtime;
    $code .=<<EXTRA;

__DATA__
Generated-on: $now
Generated-from: $self->{file}
POE::Interface: $VERSION
XML::DOM: $XML::DOM::VERSION
XML::Parser: $XML::Parser::VERSION
EXTRA
    
    return $code unless wantarray;
    return ($package.'::Lite', $code);
}

############################################################################

1;

__END__

=head1 NAME

POE::Interface - Perl extension for blah blah blah

=head1 SYNOPSIS

  use POE::Interface;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for POE::Interface was created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head1 AUTHOR

A. U. Thor, a.u.thor@a.galaxy.far.far.away

=head1 SEE ALSO

perl(1).

=cut
